/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @file	lock-detector.c
 * @brief
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>

#include "util.h"
#include "core.h"
#include "core/list.h"
#include "core/device-notifier.h"

struct lock_info {
	unsigned long hash;
	char *name;
	int state;
	int pid;
	int count;
	time_t locktime;
	time_t unlocktime;
	time_t time;
};

#define DEVICED_RUN_PATH			"/run/deviced"
#define PMLOCK_HISTORY_RUN_PATH		"/run/deviced/pmlock_detail.log"
#define PMLOCK_HISTORY_GHOST_PATH	"/var/log/ghost/boot/current/pmlock_detail.log"
#define PMLOCK_HISTORY_FILE			"/var/log/pmlock_detail.log"

#define PM_STATE_LOG_GHOST_PATH		"/var/log/ghost/boot/current/pm_state.log"
#define PM_STATE_LOG_FILE			"/var/log/pm_state.log"

#define SAVE_PMLOCK_OVER_TIME		60    /* seconds */
#define LIMIT_COUNT	128

static dd_list *lock_info_list;
static const char state_string[5][8] = {
	"START", "NORMAL", "LCDDIM", "LCDOFF", "SLEEP"
};

static time_t get_time(void)
{
	struct timespec tspec;

	clock_gettime(CLOCK_REALTIME, &tspec);

	return (time_t)tspec.tv_sec;
}

static void shrink_lock_info_list(void)
{
	dd_list *l, *l_prev;
	struct lock_info *info;
	unsigned int count;

	count = DD_LIST_LENGTH(lock_info_list);
	if (count <= LIMIT_COUNT)
		return;
	_D("list is shrink : count %d", count);

	DD_LIST_REVERSE_FOREACH_SAFE(lock_info_list, l, l_prev, info) {
		if (info->locktime == 0) {
			DD_LIST_REMOVE(lock_info_list, info);
			if (info->name)
				free(info->name);
			free(info);
			count--;
		}
		if (count <= (LIMIT_COUNT / 2))
			break;
	}
}

int set_lock_time(pid_t pid, const char *pname, int state)
{
	struct lock_info *info;
	dd_list *l;
	unsigned long val;
	char ht[15];

	if (!pname)
		return -EINVAL;

	if (state < S_NORMAL || state > S_SLEEP)
		return -EINVAL;

	save_pmlock(state, true, pid);
	snprintf(ht, sizeof(ht), "[%d]%d", state, pid);
	val = g_str_hash(ht);

	DD_LIST_FOREACH(lock_info_list, l, info) {
		if (!(info->hash == val) || !(info->state == state))
			continue;

		info->count += 1;
		if (info->locktime == 0)
			info->locktime = get_time();
		info->unlocktime = 0;
		DD_LIST_REMOVE(lock_info_list, info);
		DD_LIST_PREPEND(lock_info_list, info);

		return 0;
	}

	info = malloc(sizeof(struct lock_info));
	if (!info) {
		_E("Malloc is failed for lock_info!");
		return -ENOMEM;
	}

	info->hash = val;
	if (pid < INTERNAL_LOCK_BASE)
		info->name = strndup(pname, strlen(pname));
	else {
		char internal_case[25];

		snprintf(internal_case, sizeof(internal_case), "%s(%d)", pname, pid);
		info->name = strndup(internal_case, strlen(internal_case));
	}
	info->pid = pid;
	info->state = state;
	info->count = 1;
	info->locktime = get_time();
	info->unlocktime = 0;
	info->time = 0;

	DD_LIST_APPEND(lock_info_list, info);

	return 0;
}

static void pmlock_info_detail(char *pname, pid_t pid, time_t locktime,
		time_t unlocktime, long diff)
{
	_cleanup_fclose_ FILE *fp = NULL;
	char *lock, *unlock, *tmp;
	int ret;

	fp = fopen(PMLOCK_HISTORY_RUN_PATH, "a");
	if (fp == NULL) {
		_E("Can not open detail path");
		return;
	}

	lock = ctime(&locktime);
	tmp = strchr(lock, '\n');
	if (tmp)
		*tmp = '\0';

	if (unlocktime != 0) {
		unlock = ctime(&unlocktime);
		tmp = strchr(unlock, '\n');
		if (tmp)
			*tmp = '\0';
	} else
		unlock = "continue..";

	ret = fprintf(fp, "%6d [%24s] ~ [%24s] %5lu sec   %s \n",
			pid, lock, unlock, diff, pname);
	if (ret < 0)
		_E("write failed %d", ret);
}

int set_unlock_time(pid_t pid, int state)
{
	bool find = false;
	long diff;
	struct lock_info *info;
	dd_list *l;
	unsigned long val;
	char ht[15];

	if (state < S_NORMAL || state > S_SLEEP)
		return -EINVAL;

	save_pmlock(state, false, pid);

	snprintf(ht, sizeof(ht), "[%d]%d", state, pid);
	val = g_str_hash(ht);

	DD_LIST_FOREACH(lock_info_list, l, info) {
		if (info->hash == val && info->state == state) {
			DD_LIST_REMOVE(lock_info_list, info);
			DD_LIST_PREPEND(lock_info_list, info);
			find = true;
			break;
		}
	}

	if (!find)
		return -EINVAL;

	if (info->locktime == 0)
		return -EINVAL;

	/* update time */
	info->unlocktime = get_time();
	diff = info->unlocktime - info->locktime;
	if (diff > 0)
		info->time += diff;
	if (diff > SAVE_PMLOCK_OVER_TIME)
		pmlock_info_detail(info->name, info->pid, info->locktime, info->unlocktime, diff);

	info->locktime = 0;

	if (DD_LIST_LENGTH(lock_info_list) > LIMIT_COUNT)
		shrink_lock_info_list();

	return 0;
}

void free_lock_info_list(void)
{
	dd_list *l, *l_next;
	struct lock_info *info;

	if (!lock_info_list)
		return;

	DD_LIST_FOREACH_SAFE(lock_info_list, l, l_next, info) {
		DD_LIST_REMOVE(lock_info_list, info);
		if (info->name)
			free(info->name);
		free(info);
	}
	lock_info_list = NULL;
}

void print_lock_info_list(int fd)
{
	struct lock_info *info;
	time_t cur_time;
	dd_list *l;
	long time;
	char cur_buf[30], lock_buf[30], unlock_buf[30], buf[255];
	int ret;

	if (!lock_info_list)
		return;

	cur_time = get_time();
	ctime_r(&cur_time, cur_buf);
	cur_buf[strlen(cur_buf) - 1] = 0;
	snprintf(buf, sizeof(buf), "current time : %s\n", cur_buf);
	ret = write(fd, buf, strlen(buf));
	if (ret < 0)
		_E("write() failed (%d)", errno);

	snprintf(buf, sizeof(buf),
			"[%10s %6s] %6s      %-20s      %-20s %10s %15s\n", "hash", "state",
			"count", "last locktime", "last unlocktime", "time(s)", "process name");
	ret = write(fd, buf, strlen(buf));
	if (ret < 0)
		_E("write() failed (%d)", errno);

	DD_LIST_FOREACH(lock_info_list, l, info) {
		time = 0;
		if (info->locktime != 0 && info->unlocktime == 0)
			time = get_time() - info->locktime;

		*lock_buf = '\0';
		*unlock_buf = '\0';

		if (info->locktime) {
			ctime_r(&info->locktime, lock_buf);
			lock_buf[strlen(lock_buf) - 1] = 0;
		}
		if (info->unlocktime) {
			ctime_r(&info->unlocktime, unlock_buf);
			unlock_buf[strlen(unlock_buf) - 1] = 0;
		}

		snprintf(buf, sizeof(buf),
			"[%10lu %6s] %6d %25s %25s %10ld    %s\n",
			info->hash,
			state_string[info->state],
			info->count,
			lock_buf,
			unlock_buf,
			info->time + time,
			info->name);
		ret = write(fd, buf, strlen(buf));
		if (ret < 0)
			_E("write() failed (%d)", errno);
	}
}

static void cleanup_pmlock_statistics(void)
{
	struct lock_info *info;
	GList *n;
	time_t time;
	long diff;

	time = get_time();

	DD_LIST_FOREACH(lock_info_list, n, info) {
		if (info->unlocktime != 0)
			continue;

		diff = time - info->locktime;
		if (diff > SAVE_PMLOCK_OVER_TIME)
			pmlock_info_detail(info->name, info->pid, info->locktime, 0, diff);
	}
}

int pm_save_logdump(void)
{
	int ret;

	cleanup_pmlock_statistics();
	ret = do_copy_force(PMLOCK_HISTORY_RUN_PATH, PMLOCK_HISTORY_FILE);
	if (ret < 0)
		_W("Can not move(%d)", ret);

	save_display_log(PM_STATE_LOG_FILE);

	return ret;
}

static int pmlock_detector_poweroff_cb(void *data)
{
	int ret;

	cleanup_pmlock_statistics();
	ret = do_copy_force(PMLOCK_HISTORY_RUN_PATH, PMLOCK_HISTORY_GHOST_PATH);
	if (ret < 0)
		_W("Can not move(%d)", ret);

	save_display_log(PM_STATE_LOG_GHOST_PATH);

	return 0;
}

void pm_lock_detector_init(void)
{
	int ret;

	ret = do_mkdir(DEVICED_RUN_PATH, 0770);
	if (ret < 0) {
		_E("Fail to make directory %d", ret);
		return;
	}

	register_notifier(DEVICE_NOTIFIER_POWEROFF, pmlock_detector_poweroff_cb);
}
